import os
import sys
import shutil
import numpy as np
import yaml
import time
import matplotlib.pyplot as plt

import lica.ana_std as ana
import LICA
import lica.fileorganised as FO

import MDAnalysis as mda
import MDAnalysis.analysis.msd as msd
import MDAnalysis.analysis.rdf as Rdf
import MDAnalysis.lib.pkdtree as NLT
import MDAnalysis.lib.nsgrid as NSG
import MDAnalysis.lib.NeighborSearch as NS
import MDAnalysis.lib.distances as DIST
import MDAnalysis.lib.correlations as CORR
from scipy.stats import linregress

import MDAnalysis.analysis.diffusionmap as diffusionmap
from MDAnalysis.core.groups import AtomGroup


class MD_ANALYSIS:
	def __init__(self):
		self.cutoff = []
		self.start = []
		self.stop  = []
		self.step  = []
		self.Dir = []
		self.forcefield_itp = []
		self.tpr_file = []
		self.trr_file = []
		self.figure_extension = 0
		#self.outdir = []
		self.center_atom = []
		self.center_atom_index = []
		self.is_info_exist = False
	def get_info(self):
		finfo = FO.FILE_INFO()
		status = finfo.create_info()
		if status is False:
			return True
		elif status is True:
			self.cutoff = finfo.cutoff
			self.start = finfo.get_start()
			self.stop  = finfo.get_stop() 
			self.step  = finfo.get_step()
			self.Dir = finfo.get_DIR()
			self.forcefield_itp = finfo.get_forcefield_itp()
			self.tpr_file = finfo.get_tpr()
			self.trr_file = finfo.get_trr()
			self.figure_extension = finfo.get_figure_extension()
			#self.outdir = FO.create_dir()
			self.center_atom = finfo.get_center_atom()
			self.is_info_exist = True
			return True
		else:
			sys.stderr.write('[ERROR] something is wrong\n')
			return False
	def get_info_found(self):
		return self.is_info_exist
	def get_msd(self):
		start = time.time()
		self.elements = LICA.read_element_names(self.forcefield_itp)
		self.elements_title = []
		for e_i in self.elements:
			self.elements_title.append(e_i.title())

		if(len(self.elements)==0):
			sys.stderr.write('[ERROR] Failed at tage 1 \n')
			return False
		u = mda.Universe(self.tpr_file, self.trr_file, refresh_offsets=True)
		#u.transfer_to_memory(start = self.start, stop = self.stop , step = self.step, verbose=False)

		command = 'name '
		for ei in self.elements:
			command = command + ' ' + ei

		self.center_atom_index = -10
		for i in range(len(self.elements)):
			if (self.center_atom == self.elements[i]):
				self.center_atom_index = i
		if self.center_atom_index < 0:
			for i in range(len(self.elements)):
				if (self.center_atom == self.elements_title[i]):
					self.center_atom_index = i
		if self.center_atom_index < 0:
			sys.stderr.write('[ERROR] center_atom_index <0 \n')
			return False
		
		sys.stderr.write('commd_1 : ' + command+'\n')

		agALL = u.select_atoms(command, sorted=False)
		command = 'name '+self.elements[self.center_atom_index]

		sys.stderr.write('commd_2 : ' + command+'\n')

		ag_center_atom = u.select_atoms(command, sorted=False)

		dt = 0
		for ts in u.trajectory:
			dt = ts.dt
			break
		
		MSD = msd.EinsteinMSD(ag_center_atom, select="all", msd_type='xyz', fft=True)
		MSD.run()
		msd1 =  MSD.results.timeseries

		nframes = MSD.n_frames
		sys.stderr.write('nframes = ' + str(nframes) + '\n')

		timestep = dt # this needs to be the actual time between frames
		lagtimes = np.arange(nframes)*timestep # make the log-time axis
		fig = plt.figure()
		ax = plt.axes()
		#ax.set_yscale('log')
		#ax.set_xscale('log')
		# plot the actual MSD
		ax.plot(lagtimes, msd1, label='3D random walk')
		plt.xlabel('time (fs)')
		plt.ylabel('msd (arbitrary units)')
		#plt.show()
		if int(self.figure_extension) == 1:
			plt.savefig('msd_all_central_atoms.png')
			sys.stderr.write('check: msd_all_central_atoms.png \n')
		else:
			plt.savefig('msd_all_central_atoms.pdf')
			sys.stderr.write('check: msd_all_central_atoms.pdf \n')


		start_time = self.start
		start_index = int(start_time/timestep)
		end_index = self.stop
		linear_model = linregress(lagtimes[start_index:end_index], msd1[start_index:end_index])
		slope = linear_model.slope
		error = linear_model.rvalue
		# dim_fac is 3 as we computed a 3D msd with 'xyz'
		D = slope * 1.0/(2.0*MSD.dim_fac)
		sys.stderr.write('Average self-Diffusivity of '+self.center_atom)
		sys.stderr.write(' : {:{w}.{p}f}\n'.format(D, w=15, p=15 ) )
		end = time.time()
		sys.stdout.write('step_1, elapsed time {:{w}.{p}f} [second] \n'.format((end-start), w=5, p=5 ) )

		return True
	def get_nl(self):
		start = time.time()
		u = mda.Universe(self.tpr_file, self.trr_file)
		u.transfer_to_memory(start = self.start, stop = self.stop , step = self.step, verbose=False)

		command = 'name '
		for ei in self.elements:
			command = command + ' ' + ei

		agALL = u.select_atoms(command, sorted=False)
		command = 'name '+self.elements[self.center_atom_index]
		ag_center_atom = u.select_atoms(command, sorted=False)
		reference = ag_center_atom.positions


		self.nl_filename = 'nl.dat'
		out = open(self.nl_filename, 'w')

		out.write('cutoff_radius : {:{w}.{p}f}'.format(self.cutoff, w=5, p=2 ) )
		out.write('\n')

		for ts in u.trajectory:

			if ts.frame == 0:
				out.write('vec1 : [ ')
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[0][0], w=10, p=3 ) )
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[0][1], w=10, p=3 ) )
				out.write('{:{w}.{p}f}'.format(ts.triclinic_dimensions[0][2], w=10, p=3 ) )
				out.write(' ]\n')
				out.write('vec2 : [ ')
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[1][0], w=10, p=3 ) )
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[1][1], w=10, p=3 ) )
				out.write('{:{w}.{p}f}'.format(ts.triclinic_dimensions[1][2], w=10, p=3 ) )
				out.write(' ] \n')
				out.write('vec3 : [ ')
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[2][0], w=10, p=3 ) )
				out.write('{:{w}.{p}f},'.format(ts.triclinic_dimensions[2][1], w=10, p=3 ) )
				out.write('{:{w}.{p}f}'.format(ts.triclinic_dimensions[2][2], w=10, p=3 ) )
				out.write(' ] \n')
				out.write('atomic_environmants : \n')
			out.write('- frame : ')
			out.write('{:{w}.{p}f}'.format(ts.frame, w=5, p=0 ) )
			out.write('\n')
			out.write('  time_step : ')
			out.write('{:{w}.{p}f}'.format(ts.time, w=15, p=2 ) )
			out.write('\n')
			out.write('  dt : ')
			out.write('{:{w}.{p}f}'.format(ts.dt, w=15, p=2 ) )
			out.write('\n')
			out.write('  ATOMS : \n')
			atom_cnt = 1
			ns = NS.AtomNeighborSearch(agALL, box=ts.dimensions)
			for ca in ag_center_atom:
				dij = np.linalg.norm(ts.positions[ca.ix]-reference[atom_cnt-1])
				nl = ns.search(ca,  radius = self.cutoff)
				out.write('  - # ')
				out.write('{:{w}.{p}f}'.format(atom_cnt, w=5, p=0 ) )
				out.write('\n')
				out.write('    center_atom : ')
				out.write(ca.name)
				out.write('\n')
				out.write('    index : ')
				out.write('{:{w}.{p}f}'.format(ca.ix, w=5, p=0 ) )
				out.write('\n')
				out.write('    Coordination_number : ')
				out.write('{:{w}.{p}f}'.format(len(nl.types), w=5, p=0 ) )
				out.write('\n')
				out.write('    distance : ')
				out.write('{:{w}.{p}f}'.format(dij, w=15, p=5 ) )
				out.write('\n')
				out.write('    neighbour_types : [ ')
				for i in range(len(nl.types)):
					if i < len(nl.types)-1:
						out.write('{:>5},'.format(nl.types[i]))
					else:
						out.write('{:>5}'.format(nl.types[i]))
				out.write('    ] ')
				out.write('\n')
				out.write('    neighbour_indices : [ ')
				for i in range(len(nl.ix)):
					if i < len(nl.ix)-1:
						out.write('{:{w}.{p}f},'.format(nl.ix[i], w=10, p=0 ) )
					else:
						out.write('{:{w}.{p}f}'.format(nl.ix[i], w=10, p=0 ) )
				out.write('    ] ')
				out.write('\n')
				#print(ts.positions[nl.ix[i]])
				atom_cnt+=1
				#if atom_cnt == 10 : break
		out.close()
		sys.stderr.write('Created : ' + self.nl_filename + '\n')
		end = time.time()
		sys.stdout.write('step_2, elapsed time {:{w}.{p}f} [second] \n'.format((end-start), w=5, p=5 ) )
		return True
	def nl_file_analysis(self):
		start = time.time()
		data = ana.read_yaml_data(self.nl_filename)
		frame_properties = 'frame_properties.dat'
		out = open(frame_properties, 'w')
		for i in range(len(data[0]['atomic_environmants'])):
			frame = data[0]['atomic_environmants'][i]['frame']
			time_step = data[0]['atomic_environmants'][i]['time_step']
			dt = data[0]['atomic_environmants'][i]['dt']
			#out.write('#frame {:{w}.{p}f}'.format(frame, w=10, p=0 ))
			#out.write(' time {:{w}.{p}f}\n'.format(time_step, w=10, p=2 ))
			#for j in range(len(data[0]['atomic_environmants'][i]['ATOMS'])):
			ATOM_len = len(data[0]['atomic_environmants'][i]['ATOMS'])
			for j in range(ATOM_len):
				out.write('frame {:{w}.{p}f}'.format(frame, w=10, p=0 ))
				out.write(' time {:{w}.{p}f} '.format(time_step, w=15, p=2 ))
				out.write(' dt {:{w}.{p}f} '.format(dt, w=10, p=2 ))
				index = data[0]['atomic_environmants'][i]['ATOMS'][j]['index']
				cn = data[0]['atomic_environmants'][i]['ATOMS'][j]['Coordination_number']
				dij = data[0]['atomic_environmants'][i]['ATOMS'][j]['distance']
				out.write('  index {:{w}.{p}f}'.format(index, w=10, p=0 ))
				out.write('  cn {:{w}.{p}f}'.format(cn, w=10, p=0 ))
				out.write('  distance {:{w}.{p}f}'.format(dij, w=10, p=3 ))
				out.write('\n')
		out.close()
		sys.stderr.write('Created : ' + frame_properties + '\n')
		end = time.time()
		sys.stdout.write('step_3, elapsed time {:{w}.{p}f} [second] \n'.format((end-start), w=5, p=5 ) )

def run_analysis():

	print("MDAnalysis_version ", mda.version.__version__)

	md_ana = MD_ANALYSIS()

	ststus = md_ana.get_info()

	if ststus is False:
		return ststus
	
	if md_ana.get_info_found() is True:
		sys.stderr.write('Performing mds for all central_atoms: '+'\n')
		if md_ana.get_msd() is True:
			sys.stderr.write('Calculating neighbour_list \n')
			if md_ana.get_nl() is True:
				sys.stdout.write('Analyzing neighbour_list\n')
				md_ana.nl_file_analysis()
			else:
				return False
		else: 
			return False

	return True
